with Unchecked_Deallocation;

package body Contenedores.Dispersin_Cadenas_Separadas is
	procedure Liberar is new Unchecked_Deallocation(Nodo,P_Nodo);

	function Disp(E: in TElemento) return Natural is
	begin
		return Pseudoclave(E) mod Max_Elem;
	end Disp;
			
	procedure Buscar(C: in Tabla_Dispersin;E: in TElemento;
				 Actual, Anterior: out P_Nodo;
				 Pos: out Natural
				) is
	begin
		-- La clave debe estar en la lista encadenada indicada por su
		-- direccin de dispersin
		Pos      := Disp(E);
		Actual   := C.Elementos(Pos);
		Anterior := null;

		while Actual /= null and then Actual.Elemento /= E loop
			Anterior := Actual;
			Actual   := Actual.Siguiente;
		end loop;
	end Buscar;

	procedure Insertar(C: in out Tabla_Dispersin;E: in TElemento) is
		Actual, Anterior : P_Nodo;
		Pos: Natural;
	begin
		-- Se busca la clave y si no est se inserta en la lista indicada
		Buscar(C,E,Actual,Anterior,Pos);
		if Actual = null then
			Actual := new Nodo;
			Actual.Elemento := E;
			Actual.Siguiente := C.Elementos(Pos);
			C.Elementos(Pos) := Actual;
			C.Num_Elem := C.Num_Elem + 1;
		end if;
	end Insertar;

	procedure Extraer(C: in out Tabla_Dispersin;E: in TElemento) is
		Actual, Anterior : P_Nodo;
		Pos: Natural;
	begin
		-- Se busca la clave y si est se quita de la lista indicada
		Buscar(C,E,Actual,Anterior,Pos);
		if Actual /= null then
			if Anterior = null then
				C.Elementos(Pos) := Actual.Siguiente;
			else
				Anterior.Siguiente := Actual.Siguiente;
			end if;
			Liberar(Actual);
			C.Num_Elem := C.Num_Elem - 1;
		end if;
	end Extraer;

	procedure Vaciar(C: in out Tabla_Dispersin)  is
		Actual : P_Nodo;
	begin
		-- Hay que vaciar todas las listas encadenadas vinculadas a la tabla
		for Pos in C.Elementos'range loop
			while C.Elementos(Pos) /= null loop
				Actual := C.Elementos(Pos);
				C.Elementos(Pos) := Actual.Siguiente;
				Liberar(Actual);
			end loop;
		end loop;
		C.Num_Elem := 0;
	end Vaciar;

	function Est(C: in Tabla_Dispersin;E: in TElemento) return Boolean is
		Actual, Anterior : P_Nodo;
		Pos: Natural;
	begin
		Buscar(C,E,Actual,Anterior,Pos);
		return Actual /= null;
	end Est;

	function Tamao(C: in Tabla_Dispersin) return Natural is
	begin
		return C.Num_Elem;
	end Tamao;
end Contenedores.Dispersin_Cadenas_Separadas;

