with Ada.Numerics.Generic_Elementary_Functions;
package body Contenedores.Dispersin_Extensible is
	package Funciones is new 
		Ada.Numerics.Generic_Elementary_Functions(Float);
	use Funciones;

	function Disp(Profundidad: in Positive; E: in TElemento)
			return Fichero_Directorio.Count is
		Bits_Pseudoclave: constant Positive :=
			Integer(Float'truncation(Log(Float(Natural'Last),2.0)));
	begin
		return Fichero_Directorio.Count(
					Pseudoclave(E) /
					2 ** (Bits_Pseudoclave - Profundidad));
	end Disp;
			
	procedure Buscar(C: in Tabla_Dispersin;
					 E: in TElemento;
					 Dir: out Fichero_Directorio.Count;
					 Pos: out Natural;
					 Pos_Pag: out Fichero_Pginas.Count;
					 Una_Pgina: out Pgina;
					 Est: Out Boolean) is
	begin
		-- Se localiza la entrada del directorio para E. Es necesario sumar
		-- 2 a Disp porque el rango de valores de sta empieza en 0,
		-- mientras que el rango de direcciones de un fichero en Ada empieza
		-- en 1 y el registro 1 se est usando para guardar el n de
		-- elementos de la tabla. 
		Dir := Disp(C.Profundidad,E) + 2;
		Set_Index(C.Directorio,Dir);
		Read(C.Directorio,Pos_Pag);

		-- Se busca E en la pgina
		Set_Index(C.Pginas,Pos_Pag);
		Read(C.Pginas, Una_Pgina);
		Pos := 1;
		while Pos <= Una_Pgina.Num_Elementos and then 
	  		Una_Pgina.Elementos(Pos) /= E loop
			Pos := Pos + 1;
		end loop;
		if Pos <= Una_Pgina.Num_Elementos then
			Est := True;
		else
			Est := False;
		end if;
	end Buscar;

	procedure Insertar(C: in out Tabla_Dispersin;E: in TElemento) is
		Pos: Natural;
		Dir: Fichero_Directorio.Positive_Count;
		Pos_Pag, Pos_Pag_Aux: Fichero_Pginas.Count;
		Una_Pgina, Pgina_Div0, Pgina_Div1: Pgina;
		Elementos_Div: Natural;
		Est: Boolean;
		D1, D2: Fichero_Directorio.Count;
		N0, N1 : Natural := 0;
		Afectadas, Inicial  : Fichero_Directorio.Count;
	begin
		Buscar(C,E,Dir,Pos,Pos_Pag,Una_pgina,Est);
		if not Est then
			if Pos <= Orden then -- Cabe en la pgina
				Una_pgina.Elementos(Pos) := E;
				Una_pgina.Num_elementos := Una_Pgina.Num_Elementos + 1;
				Set_Index(C.Pginas,Pos_Pag);
				Write(C.Pginas,Una_pgina);

				C.Num_elementos := C.Num_elementos + 1;
				Set_index(C.Directorio,1);
				write(C.Directorio,Fichero_Pginas.Count(C.Num_Elementos));
			else	-- Hay que dividir
				if Una_Pgina.Profundidad = C.Profundidad then
					-- Hay que dividir el directorio
					D1 := Size(C.Directorio);
					While D1 > 1 loop
						Set_Index(C.Directorio,D1);
						Read(C.Directorio,Pos_Pag_Aux);
						D2 := (D1 - 1) * 2;

						Set_Index(C.Directorio,D2);
						Write(C.Directorio,Pos_Pag_Aux);
						Write(C.Directorio,Pos_Pag_Aux);
						D1 := D1 - 1;
					end loop;
					C.Profundidad := C.Profundidad + 1;
					Dir := (Dir - 1) * 2;
				end if;

				-- Se divide la pgina y se redistribuye su contenido
				Pgina_Div0.Profundidad := Una_Pgina.Profundidad + 1;
				Pgina_Div1.Profundidad := Una_Pgina.Profundidad + 1;
				Elementos_Div := Una_Pgina.Num_elementos;
				for i in 1 .. Elementos_Div loop
					if Disp(C.Profundidad,Una_Pgina.Elementos(I)) mod 2 = 0
					then
						N0 := N0 + 1;
						Pgina_Div0.Elementos(N0) :=
							 Una_Pgina.Elementos(I);
					else
						N1 := N1 + 1;
						Pgina_Div1.Elementos(N1) :=
							 Una_Pgina.Elementos(I);
					end if;
				end loop;

				Pgina_Div0.Num_Elementos := N0;
				Set_index(C.Pginas,Pos_Pag);
				Write(C.Pginas,Pgina_Div0);

				Pgina_Div1.Num_elementos := N1;
				Pos_Pag_Aux := Size(C.Pginas) + 1;
				Set_index(C.Pginas,Pos_Pag_Aux);
				Write(C.Pginas,Pgina_Div1);

				-- Se redireccionan las entradas del directorio
				-- afectadas por la divisin de la pgina
				Afectadas := 2 ** (C.Profundidad - Una_Pgina.Profundidad);
				Inicial   := Dir - ((Dir - 2) rem Afectadas);

				Set_Index(C.Directorio,(Inicial + (Afectadas / 2)));
				for I in (Inicial + (Afectadas / 2)) .. (Inicial + Afectadas - 1) loop
					Write(C.Directorio,Pos_Pag_Aux);
				end loop;

				-- Se reintenta la insercin
				Insertar(C,E);
			end if;
		end if;
	end Insertar;

	procedure Extraer(C: in out Tabla_Dispersin;E: in TElemento) is
		Pos: Natural;
		Dir: Fichero_Directorio.Positive_Count;
		Pos_Pag: Fichero_Pginas.Count;
		Una_Pgina: Pgina;
		Est: Boolean;

	begin
		Buscar(C,E,Dir,Pos,Pos_Pag,Una_Pgina,Est);
		if Est then
			for i in Pos .. Una_Pgina.Num_Elementos - 1 loop
				Una_Pgina.Elementos(I) := Una_Pgina.Elementos(I+1);
			end loop;
			Una_Pgina.Num_Elementos := Una_Pgina.Num_Elementos - 1;
			Set_Index(C.Pginas,Pos_Pag);
			Write(C.Pginas,Una_Pgina);

			C.Num_Elementos := C.Num_Elementos - 1;
			Set_Index(C.Directorio,1);
			Write(C.Directorio,Fichero_Pginas.Count(C.Num_Elementos));
		end if;
	end Extraer;

	procedure Vaciar(C: in out Tabla_Dispersin) is
	begin
		Crear(C,C.Ruta);
	end Vaciar;

	function Est(C: in Tabla_Dispersin;E: in TElemento) return boolean is
		Pos: Natural;
		Dir: Fichero_Directorio.Positive_Count;
		Pos_Pag : Fichero_Pginas.Count;
		Una_Pgina: Pgina;
		Est: Boolean;
	begin
		Buscar(C,E,Dir,Pos,Pos_Pag,Una_Pgina,Est);
		return Est;
	end Est;

	function Es_Vaco(C: in Tabla_Dispersin) return Boolean is
	begin
		return Tamao(C) = 0;
	end Es_Vaco;

	function Tamao(C: in Tabla_Dispersin) return Natural is
	begin
		return C.Num_Elementos;
	end Tamao;

	procedure Abrir(C: in out Tabla_Dispersin;Ruta: in Unbounded_String) is
	begin
		if Is_Open(C.Directorio) then
			Cerrar(C);
		end if;

		Open(C.Directorio,Inout_File,To_String(Ruta & "Directorio.dsp"));
		Open(C.Pginas,Inout_File,To_String(Ruta & "Datos.dsp"));
      Read(C.Directorio, Num_Elementos);
      C.Num_Elementos := Natural(Num_Elementos);
		C.Profundidad := Positive(Float'truncation(
						Log(Float(Size(C.Directorio)-1),2.0)));
		C.Ruta := Ruta;
	end Abrir;

	procedure Crear(C: in out Tabla_Dispersin;Ruta: in Unbounded_String) is
		Una_Pgina: Pgina;
	begin
		if Is_Open(C.Directorio) then
			Cerrar(C);
		end if;
		-- Se crea el directorio inicial
		Create(C.Directorio,Name => To_String(Ruta & "Directorio.dsp"));
		Write(C.Directorio,0); -- N total de elementos almacenados
		Write(C.Directorio,1); -- 1 entrada del directorio
		Write(C.Directorio,1); -- 2 entrada del directorio

		-- Se crea la pgina inicial
		Una_Pgina.Profundidad := 0;
		Una_Pgina.Num_elementos := 0;
		Create(C.Pginas,Name => To_String(Ruta & "Datos.dsp"));
		Write(C.Pginas,Una_Pgina);

		-- Se inicializan los datos internos del tad
		C.Profundidad := 1;
		C.Num_Elementos := 0;
		C.Ruta := Ruta;
	end Crear;

	procedure Cerrar(C: in out Tabla_Dispersin) is
	begin
		Close(C.Directorio);
		Close(C.Pginas);
	end Cerrar;
end Contenedores.Dispersin_Extensible;
