with Ada.Strings.Unbounded, Direct_Io;
use  Ada.strings.Unbounded;
generic
	Orden: Positive; --Nmero de entradas por pgina
	with function Pseudoclave(E: in TElemento) return Natural;
package Contenedores.Dispersin_Extensible is
	type Tabla_Dispersin is new Contenedor with private;

	--Operaciones heredadas de Contenedores
	procedure Insertar(C: in out Tabla_Dispersin;E: in TElemento);
	procedure Extraer(C: in out Tabla_Dispersin;E: in TElemento);
	procedure Vaciar(C: in out Tabla_Dispersin);
	function Est(C: in Tabla_Dispersin;E: in TElemento) return Boolean;
	function Tamao(C: in Tabla_Dispersin) return Natural;

	---------------------------------------------
	-- Operaciones para gestionar los ficheros --
	-- asociados a la tabla de dispersin      --
	---------------------------------------------
	procedure Abrir(C: in out Tabla_Dispersin; Ruta: in Unbounded_String);
		-- Abre una tabla de dispersin existente en la ubicacin indicada
		-- por Ruta
	procedure Crear(C: in out Tabla_Dispersin; Ruta: in Unbounded_String);
		-- Crea una tabla de dispersin en la ubicacin indicada
		-- por Ruta
	procedure Cerrar(C: in out Tabla_Dispersin);
		-- Cierra los ficheros asociados a la tabla de dispersin

private
	-- Estructura de una pgina de la tabla
	type Almacen is array(1..Orden) of TElemento;

	type Pgina is record
		Profundidad: Natural;
		Elementos  : Almacen;
		Num_Elementos: Natural;
	end record;

	package Fichero_Pginas is new Direct_Io(Pgina);
	use Fichero_Pginas;

	-- El directorio contiene nmeros de posiciones de pginas,
	-- (salvo el primer nmero, que indica total de elementos)
	package Fichero_directorio is new Direct_Io(Fichero_Pginas.Count);
	use Fichero_Directorio;

	-- Estructura interna para gestionar la tabla
	type Tabla_Dispersin is new Contenedor with record
		Ruta: Unbounded_String;	-- Ubicacin de los ficheros
		Directorio: Fichero_Directorio.File_Type; -- Fichero directorio
		Pginas   : Fichero_Pginas.File_Type;	  -- Fichero de datos
		Profundidad: Positive;
		Num_Elementos: Natural;
    end record;
end Contenedores.Dispersin_Extensible;
