#ifndef INC_DISPERSION_EXTENSIBLE
#define INC_DISPERSION_EXTENSIBLE
#include "..\comun\fichero.h"
 #include "..\comun\contenedor.h"
template <class T, int TDC=10>
   class DispersionExtensible: public Contenedor<T>{
      static const char *extensionFicheroDirectorio;
      static const char *extensionFicheroDatos;
      struct Pagina{
         int direccion;   //Direccion de la pgina en el fichero
         int profundidad; //bits en la funcin de dispersin que coinciden
         int numEle;
         T elementos[TDC];
         Pagina(){
            direccion=-1;
            profundidad=0;
            numEle=0;}
      };
      Fichero<int> directorio;
      FicheroAvanzado<Pagina> paginas;
      int  numEle;
      int  profundidad;
      string camino;
      void buscar(const T &e, Pagina &pagina, int &posDir, int & pos, bool &esta) const;
      int  log2(unsigned int dato) const{
         int res=0;
         for(int i=1; i<32; i++)
            if(((dato>>i) & 1) == 1) 
               res=i;
         return res;
      }
      int dispersion(int profundidad, T e) const{
         int pse=pseudoClave(e);
         int res=0;
         for(int i=0; i<profundidad; i++){
            res<<=1;
            if(pse & 1 == 1)
               res|=1;
            pse >>=1;
         }
         return res;
      }
      Pagina leerPagina(int dir);
      void escribirPagina(const Pagina &);
   public:
      DispersionExtensible(){
         numEle=0;
         profundidad=0;
      }
      void crear(const char *ruta);
      void abrir(const char *ruta);
      void cerrar();
      void insertar(const T&);
      void extraer(const T&);
      bool encuentra(const T&) const;
      void vaciar();
      int  cardinal() const;
   };
template <class T, int TDC>
   const char *DispersionExtensible<T,TDC>::extensionFicheroDirectorio="_directorio.hashExtensible";
template <class T, int TDC>
   const char *DispersionExtensible<T,TDC>::extensionFicheroDatos="_datos.hashExtensible";

template <class T, int TDC>
   typename DispersionExtensible<T,TDC>::Pagina DispersionExtensible<T,TDC>::leerPagina(int dir){
      Pagina pagina=paginas.leer(dir);
      if(pagina.direccion != dir) 
         throw excepcionFichero("Error al leer pgina");
   }

template <class T, int TDC>
   void DispersionExtensible<T,TDC>::escribirPagina(const Pagina &pagina){
      paginas.escribir(pagina,pagina.direccion);
   }

template <class T, int TDC>
   void DispersionExtensible<T,TDC>::crear(const char *ruta){
      camino=ruta;
      directorio.crear(camino+extensionFicheroDirectorio);
      paginas.crear(camino+extensionFicheroDatos,1);
      Pagina pagina;  //Pgina vaca
      pagina.direccion=0;
      escribirPagina(pagina);
      profundidad=1;
      numEle=0;
      directorio.escribir(pagina.direccion,0); //Entrada 0 del directorio apunta a la pgina 0
      directorio.escribir(pagina.direccion,1); //Entrada 1 del directorio apunta a la pgina 0
      paginas.adjunto(0,numEle);//Nmero de elementos en el contenedor 0
   }
template <class T, int TDC>
   void DispersionExtensible<T,TDC>::abrir(const char *ruta){
      camino=ruta;
      directorio.abrir(camino+extensionFicheroDirectorio);
      paginas.abrir(camino+extensionFicheroDatos);
      numEle=paginas.adjunto(0);
      profundidad = log2(directorio.tamano());
   }
template <class T, int TDC>
   void DispersionExtensible<T,TDC>::cerrar(){
      directorio.cerrar();
      paginas.cerrar();
   }
template <class T, int TDC>
   void DispersionExtensible<T,TDC>::buscar(const T &e, Pagina &pagina, int &posDir, int & pos, bool &esta)const{
      int posPagina;
      posDir=dispersion(profundidad,e);
      posPagina=directorio.leer(posDir);
      pagina=paginas.leer(posPagina);
      esta=false;
      for(pos=0; pos<pagina.numEle; pos++)
         if(pagina.elementos[pos]==e){
            esta=true;
            break;
         }
   }

template <class T, int TDC>
   void DispersionExtensible<T,TDC>::insertar(const T &e){
      int posDir, pos;
      bool esta;
      Pagina pagina;
      buscar(e,pagina,posDir,pos,esta);
      if(!esta){
         if(pagina.numEle < TDC) {//Cabe en la pgina
            pagina.elementos[pagina.numEle++] = e;
            escribirPagina(pagina);
            numEle++;
            paginas.adjunto(0,numEle);
         }
         else { //Hay que dividir la pgina
            if(pagina.profundidad == profundidad){ //Hay que dividir el directorio
               int nentradas=directorio.tamano();
               for(int i=nentradas; i<nentradas*2; i++) //Rellenamos [0..nentradas-1]=>[0..(nentradas*2)-1]
                  directorio.escribir(i,i);
               for(int i=nentradas-1; i>=0; i--){ //Duplicamos. cada i => 2*i y 2*i+1
                  int dato=directorio.leer(i);
                  directorio.escribir(dato,i*2);
                  directorio.escribir(dato,i*2+1);
               }
               profundidad++;
               posDir= 2*posDir;
            }
            Pagina paginaDivision0, paginaDivision1;
            paginaDivision0.profundidad= pagina.profundidad+1;
            paginaDivision0.direccion= pagina.direccion; //En la misma pgina los de 0
            paginaDivision1.profundidad= pagina.profundidad+1;
            paginaDivision1.direccion= paginas.tomarPagina(); //En una nueva los de 1
            for(int i=0; i<pagina.numEle; i++)
               if ((dispersion(pagina.profundidad+1,pagina.elementos[i])&1) == 0)
                  paginaDivision0.elementos[paginaDivision0.numEle++]=pagina.elementos[i];
               else
                  paginaDivision1.elementos[paginaDivision1.numEle++]=pagina.elementos[i];
            escribirPagina(paginaDivision0);
            escribirPagina(paginaDivision1);
            int difProfundidad=profundidad-paginaDivision0.profundidad;
            int nEntradasDir=1<<difProfundidad;
            int posDirInicial  =((posDir>>(difProfundidad+1))<<(difProfundidad+1))+nEntradasDir;
            int posDirFinal    = posDirInicial+ nEntradasDir;
            for(int i=posDirInicial; i<posDirFinal; i++)
               directorio.escribir(paginaDivision1.direccion,i);
            insertar(e); //Se reintenta la insercin
         }
      }
   }

template <class T, int TDC>
   void DispersionExtensible<T,TDC>::extraer(const T &e){
      int posDir, pos;
      bool esta;
      Pagina pagina;
      buscar(e,pagina,posDir,pos,esta);
      if(esta){
         pagina.numEle--;
         for(int i=pos; i<pagina.numEle; i++)
            pagina.elementos[i]=pagina.elementos[i+1];
         escribirPagina(pagina);
         numEle--;
         paginas.adjunto(0,numEle);
      }
   }

template <class T, int TDC>
   void DispersionExtensible<T,TDC>::vaciar(){
      if(directorio.abierto()){
         cerrar();
         crear(camino.c_str());
      }
   }

template <class T, int TDC>
   bool DispersionExtensible<T,TDC>::encuentra(const T &e)const {
      int posDir, pos;
      bool esta;
      Pagina pagina;
      buscar(e,pagina,posDir,pos,esta);
      return esta;
   }

template <class T, int TDC>
   int DispersionExtensible<T,TDC>::cardinal() const{
      return numEle;
   }
#endif
