package dispersinExtensible;

import comn.*;
import contenedores.Almacenable;
import contenedores.ContenedorPersistente;

/**
 * Implementacin de un contenedor utilizando dispersin extensible
 * basada en directorio
 */
public class DispersinExtensible implements ContenedorPersistente {
   static final String extFicheroDirectorio = ".dir";
   static final String extFicheroDatos = ".dat";
   int TDC; //Tamao mximo de celda
   int tamaoDatos; //Tamao en bytes de los datos almacenados

   class Pgina {
      public int direccin;      //Direccion de la pgina en el fichero
      public int profundidad;    //bits coincidentes en f. dispersin
      public int numEle;         //Datos ocupados en la pgina
      public byte[][] elementos; //Vector de datos en la pgina

      Pgina() {
         direccin = -1;
         profundidad = 0;
         numEle = 0;
         elementos = new byte[TDC][];
      }

      int tamao() { //Tamao en bytes de la pgina
         int tam = 3 * Conversor.INTBYTES; //Cantidad base
         tam += TDC * tamaoDatos; //Se aaden los datos
         return tam;
      }

      /**
       * @return un byte[] que representa la pgina
       */
      byte[] aByte() {
         int tam = tamao();
         byte[] res = new byte[tam];
         int pos = 0;
         pos = Conversor.aade(res, Conversor.aByte(direccin), pos);
         pos = Conversor.aade(res, Conversor.aByte(profundidad), pos);
         pos = Conversor.aade(res, Conversor.aByte(numEle), pos);
         for (int i = 0; i < numEle; i++)
            pos = Conversor.aade(res, elementos[i], pos);
         return res;
      }

      /**
       * Restaura el objeto actual partiendo de
       * un byte[] obtenido de aByte()
       */
      void deByte(byte[] datos) {
         int lbi=Conversor.INTBYTES; //longitud en bytes de los enteros
         direccin = Conversor.aInt(Conversor.toma(datos, 0, lbi));
         profundidad = Conversor.aInt(Conversor.toma(datos, lbi, lbi));
         numEle = Conversor.aInt(Conversor.toma(datos, lbi * 2, lbi));
         for (int i = 0; i < numEle; i++){
            int pos = lbi * 3 + i * tamaoDatos;
            elementos[i] = Conversor.toma(datos, pos, tamaoDatos);
         }
      }
   };

   FicheroAvanzado directorio; //Fichero con el directorio
   FicheroAvanzado pginas;    //Fichero con las pginas
   int numEle;                 //Nmero de datos almacenados
   int profundidad; //Tamao del directorio 2^profundidad
   String camino;   //Nombre base de los ficheros (falta extensin)

   int log2(long dato) {
      int res = 0;
      for (int i = 1; i < 32; i++)
         if (((dato >> i) & 1) == 1) res = i;
      return res;
   }

   /**
    * Funcin de hash que devuelve tantos bits como indica
    * profundidad menos significativos y en orden inverso
    * al de la pseudoclave
    */
   protected int dispersin(int profundidad, Almacenable e) {
      int pse = e.hashCode();
      int res = 0;
      for (int i = 0; i < profundidad; i++) {
         res <<= 1;
         if ((pse & 1) == 1) res |= 1;
         pse >>= 1;
      }
      return res;
   }

   /**
    * Lee y devuelve la pgina en la posicin dir
    * del fichero de datos.
    */
   Pgina leerPgina(int dir) {
      Pgina pgina = new Pgina();
      pgina.deByte(pginas.leer(dir));
      if (pgina.direccin != dir) 
    	  throw new Fichero.ExcepcionFichero("Error al leer pgina");
      return pgina;
   }

   /**
    * Escribe la pgina en la posicin correspondiente
    * en el fichero de datos.
    */
   void escribirPgina(Pgina pgina) {
      pginas.escribir(pgina.aByte(), pgina.direccin);
   }
 
   /**
    * Busca el elemento devolviendo su informacin posicional
    */
   int buscar(Almacenable e, Pgina pgina, int posDir) {
      int pos;
      int posPgina;
      posPgina = Conversor.aInt(directorio.leer(posDir));
      pgina.deByte(pginas.leer(posPgina));
      for (pos = 0; pos < pgina.numEle; pos++)
         if (e.equals(e.deByte(pgina.elementos[pos]))) { return pos; }
      return -1;
   }

   public DispersinExtensible() {
      numEle = 0;
      profundidad = 0;
      directorio = new FicheroAvanzado();
      pginas = new FicheroAvanzado();
   }

   /**
    * Crea el directorio y el fichero de datos inicializando
    * el directorio a dos entradas que referencian una misma
    * pgina vaca
    */
   public void crear(String ruta, int tamaoDatos, int TDC) {
      this.tamaoDatos = tamaoDatos;
      this.TDC = TDC;
      Pgina pagina = new Pgina(); //Pgina vaca
      camino = ruta;
      directorio.crear(camino + extFicheroDirectorio, Conversor.INTBYTES);
      pginas.crear(camino + extFicheroDatos, pagina.tamao(), 3);
      pagina.direccin = pginas.tomarPgina();
      escribirPgina(pagina);
      profundidad = 1;
      numEle = 0;
      directorio.escribir(Conversor.aByte(pagina.direccin), 0);
      //Entrada 0 del directorio apunta a la pgina 0
      directorio.escribir(Conversor.aByte(pagina.direccin), 1);
      //Entrada 1 del directorio apunta a la pgina 0
      pginas.adjunto(0, numEle); //Nmero de elementos en el contenedor 0
      pginas.adjunto(1, tamaoDatos);//Nmero de bytes en cada dato
      pginas.adjunto(2, TDC); //Nmero de entradas mximas por pgina
   }

   public void crear(String nombre, int tamaoDatos) {
      crear(nombre, tamaoDatos, 100);
   }

   /**
    * Abre los ficheros. Calcula profundidad y lee el nmero de datos.
    */
   public void abrir(String ruta) {
      camino = ruta;
      directorio.abrir(camino + extFicheroDirectorio);
      pginas.abrir(camino + extFicheroDatos);
      numEle = pginas.adjunto(0);
      tamaoDatos = pginas.adjunto(1);
      TDC = pginas.adjunto(2);
      profundidad = log2(directorio.tamao());
   }

   public void vaciar() {
      if (directorio.abierto()) {
         cerrar();
         crear(camino, tamaoDatos, TDC);
      }
   }

   public void cerrar() {
      directorio.cerrar();
      pginas.cerrar();
   }

   public void insertar(Object o) {
      Almacenable alma = (Almacenable) o;
      byte[] e = alma.aByte();
      int posDir = dispersin(profundidad, alma);
      Pgina pgina = new Pgina();
      int pos = buscar(alma, pgina, posDir);
      if (pos < 0) {
         if (pgina.numEle < TDC) {//Cabe en la pgina
            pgina.elementos[pgina.numEle++] = e;
            escribirPgina(pgina);
            numEle++;
            pginas.adjunto(0, numEle);
         } else { //Hay que dividir la pgina
            if (pgina.profundidad == profundidad) { //Hay que dividir el
               // directorio
               int nentradas = (int) directorio.tamao();
               for (int i = nentradas; i < nentradas * 2; i++)
                  //Rellenamos [0..nentradas-1]=>[0..(nentradas*2)-1]
                  directorio.escribir(Conversor.aByte(i), i);
               for (int i = nentradas - 1; i >= 0; i--) { //Duplicamos.
                  // cada i => 2*i
                  // y 2*i+1
                  byte[] dato = directorio.leer(i);
                  directorio.escribir(dato, i * 2);
                  directorio.escribir(dato, i * 2 + 1);
               }
               profundidad++;
               posDir = 2 * posDir;
            }
            Pgina pginaDivisin0 = new Pgina();
            Pgina pginaDivisin1 = new Pgina();
            pginaDivisin0.profundidad = pgina.profundidad + 1;
            pginaDivisin0.direccin = pgina.direccin;
            //En la misma pgina los de 0
            pginaDivisin1.profundidad = pgina.profundidad + 1;
            pginaDivisin1.direccin = pginas.tomarPgina();
            //En una nueva los de 1
            for (int i = 0; i < pgina.numEle; i++){
               Almacenable dato = alma.deByte(pgina.elementos[i]);
               int valorDispersin=dispersin(pgina.profundidad + 1, dato);
               if ((valorDispersin & 1) == 0)
                    pginaDivisin0.elementos[pginaDivisin0.numEle++] = pgina.elementos[i];
               else pginaDivisin1.elementos[pginaDivisin1.numEle++] = pgina.elementos[i];
            }
            escribirPgina(pginaDivisin0);
            escribirPgina(pginaDivisin1);
            int difProfundidad = profundidad - pginaDivisin0.profundidad;
            int nEntradasDir = 1 << difProfundidad; // 2^difProfundidad
            int posDirInicial = posDir- (posDir % (nEntradasDir*2))+ nEntradasDir; 
            int posDirFinal = posDirInicial + nEntradasDir;
            byte[] dirByte = Conversor.aByte(pginaDivisin1.direccin);
            for (int i = posDirInicial; i < posDirFinal; i++)
               directorio.escribir(dirByte, i);
            insertar(o); //Se reintenta la insercin
         }
      }
   }

   public void extraer(Object o) {
      Almacenable alma = (Almacenable) o;
      int posDir = dispersin(profundidad, alma);
      Pgina pgina = new Pgina();
      int pos = buscar(alma, pgina, posDir);
      if (pos >= 0) {
         pgina.numEle--;
         for (int i = pos; i < pgina.numEle; i++)
            pgina.elementos[i] = pgina.elementos[i + 1];
         escribirPgina(pgina);
         numEle--;
         pginas.adjunto(0, numEle);
      }
   }

   public boolean est(Object o) {
      Almacenable alma = (Almacenable) o;
      int posDir = dispersin(profundidad, alma);
      Pgina pgina = new Pgina();
      return buscar(alma, pgina, posDir) >= 0;
   }

   public boolean esVaco() {
      return tamao() == 0;
   }

   public int tamao() {
      return numEle;
   }
}