package pruebaLineal;
import contenedores.*;
/**
 * Implementacin de un contenedor usando una tabla de dispersin
 * con prueba lineal.
 */
public class TablaDispersin implements Contenedor { 
	Object elementos[];
	boolean ocupados[];
	int numElementos = 0;

	/**
	 * Se inicializa la tabla con capacidad por defecto para
	 * 51 elementos
	 */
	public TablaDispersin() {
	    this(51);
	}

	/**
	 * Se inicializa la tabla con la capacidad indicada
	 */
	public TablaDispersin(int capacidad) {
		elementos = new Object[capacidad];
		ocupados = new boolean[capacidad];
		for(int i = 0; i < ocupados.length; i ++)
			ocupados[i] = false;
	}

	/**
	 * Hace corresponder la pseudoclave de hash del (hashCode) objeto
	 * con una posicin del espacio de almacenamiento (elementos)
	 */
	private int disp(Object e) {
		return e.hashCode() % elementos.length;
	}

	/**
	 * Devuelve la posicin de e en la tabla de dispersin;
	 * devuelve -1 sino lo encuentra
	 */
	private int buscar(Object e) {
		int actual = disp(e);
		int lmite = actual;
		// La clave se busca empezando en la posicin que le corresponde
		// y continuando hasta encontrarla, hallar una posicin vaca o
		// volver al punto de partida
		do {
			if (!ocupados[actual]) return actual;
			else
				if (elementos[actual].hashCode() == e.hashCode())
					return actual;
			actual = (actual + 1) % elementos.length;
		} while (actual != lmite);
		return -1;
	}

	/* (non-Javadoc)
	 * @see contenedores.Contenedor#insertar(java.lang.Object)
	 */
	public void insertar(Object e) {
		// Se localiza la posicin en que est o debe estar la
		// clave y se inserta si hay hueco (si no se produce la
		// excepcin ArrayIndexOutOfBoundsException)
		int actual = buscar(e);
		if (!ocupados[actual]) {
			elementos[actual] = e;
			ocupados[actual]  = true;
			numElementos++;
		}
	}
	
	/* (non-Javadoc)
	 * @see contenedores.Contenedor#extraer(java.lang.Object)
	 */
	public void extraer (Object e){
		int actual = buscar(e);
		if ((actual != -1) &&  (ocupados[actual])) {
			// Si se encontr, se quita dejando libre la posicin
			ocupados[actual] = false;
			numElementos--;
			
			// Hay que tapar el hueco reubicando las claves posteriores
			// segn la secuencia de bsqueda
			int lmite = actual;
			int siguiente = (actual + 1) % elementos.length;
			do {
				if (!ocupados[siguiente]) break;
				
				int dondeVa = disp(elementos[siguiente]);
				if (actual < siguiente) {
					// La zona prohibida est comprendida entre actual
					// y siguiente
					if ((dondeVa <= actual) || (dondeVa > siguiente)) {
						elementos[actual]  = elementos[siguiente];
						ocupados[actual]   = true;
						ocupados[siguiente]= false;
						actual = siguiente;
					}
				}
				else {
					// La zona prohibida abarca desde siguiente hasta
					// el final y desde el inicio hasta actual.
					if ((dondeVa <= actual) && (dondeVa > siguiente)) {
						elementos[actual]  = elementos[siguiente];
						ocupados[actual]   = true;
						ocupados[siguiente]= false;
						actual = siguiente;
					}
				}
				siguiente =(siguiente + 1) % elementos.length;
			} while (siguiente != lmite);
		}
	}
		
	/* (non-Javadoc)
	 * @see contenedores.Contenedor#vaciar()
	 */
	public void vaciar() {
		for (int i = 0; i < ocupados.length; i++)
			ocupados[i] = false;
		numElementos = 0;			
	}

	/* (non-Javadoc)
	 * @see contenedores.Contenedor#est(java.lang.Object)
	 */
	public boolean est(Object e) {
		int actual = buscar(e);
		return (actual != -1)?ocupados[actual]:false; 
	}
	
	/* (non-Javadoc)
	 * @see contenedores.Contenedor#esVaco()
	 */
	public boolean esVaco() {
		return (numElementos == 0);
	}
	
	/* (non-Javadoc)
	 * @see contenedores.Contenedor#tamao()
	 */
	public int tamao() {
		return numElementos;
	}
}
